use nom::bytes::complete::take_while1;
use nom::character::complete::space0;
use nom::error::{context, ErrorKind};
use nom::multi::separated_list0;
use nom::sequence::{delimited, separated_pair};
use nom::{
	branch::alt,
	bytes::complete::{is_not, tag},
	combinator::map,
	multi::many0,
	sequence::preceded,
	IResult, InputTake,
};
use std::fmt::Display;

/// Holds one part of the value of an environment variable. Static parts will be used verbatim,
/// while variables will be looked up at render time when an environment file is applied.
/// Comments are always ignored by the renderer, but are preserved for serialization.
#[derive(Clone, Debug, PartialEq)]
pub enum ValuePart {
	/// A static value that will be rendered verbatim.
	Static(String),
	/// The exact name of an environment variable that will be looked up at render time.
	/// Case-sensitivity is dependent on the operating system.
	Variable(String),
	/// Arbitrary text that will not be added to the environment.
	Comment(String),
}

impl Display for ValuePart {
	fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
		match self {
			Self::Static(static_value) => write!(f, "{}", static_value),
			Self::Variable(variable_name) => write!(f, "${{{}}}", variable_name),
			Self::Comment(comment_value) => write!(f, "#{}", comment_value),
		}
	}
}

impl ValuePart {
	/// Create a static value that will be used verbatim.
	pub fn new(value: impl ToString) -> Self {
		Self::Static(value.to_string())
	}
	/// Create a reference to another environment variable that will be interpolated into the value.
	pub fn variable(value: impl ToString) -> Self {
		Self::Variable(value.to_string())
	}
	/// Create a comment that will be ignored by the value renderer.
	pub fn comment(value: impl ToString) -> Self {
		Self::Comment(value.to_string())
	}

	fn var_from_span(span: Span) -> Self {
		Self::Variable(span.fragment().to_string())
	}

	fn static_from_span(span: Span) -> Self {
		Self::Static(span.fragment().to_string())
	}

	#[allow(dead_code)]
	fn static_from_string(span: String) -> Self {
		Self::Static(span)
	}

	fn comment_from_span(span: Span) -> Self {
		Self::Comment(span.fragment().to_string())
	}
}

/// A whole line of an environment file. May contain comments, static strings, or interpolation
/// variables.
#[derive(Clone, Debug, PartialEq)]
pub enum FileLine {
	/// A blank lin in a file, containing only whitespace. Ignored for all purposes except 1-1
	/// serialisation.
	Empty,
	/// A full line comment, starting with a '#' and containing arbitrary text for the rest of the
	/// line.
	Comment(String),
	/// A key-value pair, where the key is a string and the value is a combination of static parts,
	/// dynamic parts, and comments.
	KeyValue { key: String, value: Vec<ValuePart> },
}

impl Display for FileLine {
	fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
		match self {
			Self::Empty => write!(f, ""),
			Self::Comment(comment) => write!(f, "#{}", comment),
			Self::KeyValue { key, value } => {
				write!(f, "{}={}", key, value.iter().map(|part| part.to_string()).collect::<String>())
			}
		}
	}
}

impl FileLine {
	/// Create a new empty line.
	pub fn empty() -> Self {
		Self::Empty
	}

	/// Create a new full line comment.
	pub fn comment(comment: impl ToString) -> Self {
		Self::Comment(comment.to_string())
	}

	/// Create a new key-value line, where the value may be a combination of static strings,
	/// variable references, and comments.
	///
	/// Lines constructed this way may contain comments in the middle of strings and references
	/// without issue, but will not be able to correctly serialise if writing the non-compliant
	/// line to a file.
	pub fn key_value(key: impl ToString, value: impl ToOwned<Owned = Vec<ValuePart>>) -> Self {
		Self::KeyValue {
			key: key.to_string(),
			value: value.to_owned(),
		}
	}

	fn empty_from_span(_span: Span) -> Self {
		Self::Empty
	}
	fn key_value_form_span_pair((key_span, value): (Span, Vec<ValuePart>)) -> Self {
		Self::KeyValue {
			key: key_span.fragment().to_string(),
			value,
		}
	}
	fn comment_from_span(span: Span) -> Self {
		Self::Comment(span.fragment().to_string())
	}

	/// Returns true if this line can be evaluated by itself. A line cannot be evaluated by itself
	/// if it references other environment variables.
	pub fn is_complete(&self) -> bool {
		match self {
			Self::KeyValue { value, .. } => value.iter().all(|part| !matches!(part, &ValuePart::Variable(..))),
			_ => true,
		}
	}

	/// Returns a copy of this line without comments. If the entire line is a comment, it is
	/// converted to `Line::Empty`. Otherwise, comment parts of key-value lines are removed.
	pub fn strip_comments(&self) -> Self {
		match self {
			Self::Empty => Self::Empty,
			Self::Comment(_) => Self::Empty,
			Self::KeyValue { key, value } => Self::KeyValue {
				key: key.clone(),
				value: value
					.iter()
					.filter(|part| !matches!(part, ValuePart::Comment(..)))
					.cloned()
					.collect(),
			},
		}
	}

	/// Convert the line into a complete value string, interpolating dynamic variables and stripping
	/// comments. Empty lines or whole line comments will return an empty string.
	pub fn assemble_value(&self) -> String {
		match self {
			Self::Empty => String::new(),
			Self::Comment(_) => String::new(),
			Self::KeyValue { value, .. } => value
				.iter()
				.map(|part| match part {
					ValuePart::Static(val) => val.clone(),
					ValuePart::Variable(var) => std::env::var(var).unwrap_or_default(),
					ValuePart::Comment(_) => String::new(),
				})
				.collect::<String>(),
		}
	}
}

type Span<'a> = nom_locate::LocatedSpan<&'a str>;

/// Takes everything until the end of line.
fn take_until_eol(input: Span) -> IResult<Span, Span> {
	if input.fragment().starts_with('\n') {
		return Ok((input, unsafe {
			Span::new_from_raw_offset(input.location_offset(), input.location_line(), "", ())
		}));
	}
	is_not("\n")(input)
}

fn is_ident_first_char(c: char) -> bool {
	c.is_alphabetic() || c == '_'
}
fn is_ident_any_char(c: char) -> bool {
	c.is_alphanumeric() || c == '_'
}

fn identifier(input: Span) -> IResult<Span, Span> {
	let mut last_index = 0;
	let mut input_chars = input.chars().peekable();
	let mut valid = match input_chars.next() {
		Some(c) => is_ident_first_char(c),
		None => false,
	};

	'checkloop: while valid {
		last_index += 1;
		let next_char = input_chars.next();
		if next_char.is_none() {
			break 'checkloop;
		}
		valid = next_char.map(is_ident_any_char).unwrap_or(false);
	}

	if last_index == 0 {
		Err(nom::Err::Error(nom::error::Error::new(input, ErrorKind::Fail)))
	} else {
		Ok(input.take_split(last_index))
	}
}

fn value_separator(input: Span) -> IResult<Span, Span> {
	context("Value Separator", delimited(space0, tag("="), space0))(input)
}

fn values(input: Span) -> IResult<Span, Vec<ValuePart>> {
	context(
		"Environment Variable Values",
		many0(alt((
			preceded(tag("#"), map(take_until_eol, ValuePart::comment_from_span)),
			preceded(tag("$"), map(identifier, ValuePart::var_from_span)),
			delimited(tag("${"), map(identifier, ValuePart::var_from_span), tag("}")),
			map(
				take_while1(|c: char| c != '$' && c != '\n' && c != '#'),
				ValuePart::static_from_span,
			),
		))),
	)(input)
}

fn line(input: Span) -> IResult<Span, FileLine> {
	context(
		"Environment Variable Line",
		delimited(
			space0,
			alt((
				map(preceded(tag("#"), take_until_eol), FileLine::comment_from_span),
				map(
					separated_pair(identifier, value_separator, values),
					FileLine::key_value_form_span_pair,
				),
				map(take_until_eol, FileLine::empty_from_span), // Empty line
			)),
			space0,
		),
	)(input)
}

pub(crate) fn file(input: Span) -> IResult<Span, Vec<FileLine>> {
	context("Environment File", separated_list0(tag("\n"), line))(input)
}

#[cfg(test)]
mod tests {
	use super::*;

	#[test]
	fn test_take_until_eol_empty_line() {
		let input = Span::new(" \n");
		let result = take_until_eol(input);
		let (remaining, parsed) = result.expect("Failed to parse");

		assert_eq!(parsed.fragment(), &" ");
		assert_eq!(remaining.fragment(), &"\n");
	}

	#[test]
	fn test_take_until_eol_whitespace_only() {
		let input = Span::new("    \t  ");
		let result = take_until_eol(input);
		let (remaining, parsed) = result.expect("Failed to parse");

		assert_eq!(parsed.fragment(), &"    \t  ");
		assert_eq!(remaining.fragment(), &"");
	}

	#[test]
	fn test_identifier_alphanumeric() {
		let input = Span::new("abc123_XYZ");
		let result = identifier(input);
		let (remaining, parsed) = result.expect("Failed to parse identifier");

		assert_eq!(parsed.fragment(), &"abc123_XYZ");
		assert_eq!(remaining.fragment(), &"");
	}

	#[test]
	fn test_identifier_starting_with_number() {
		let input = Span::new("123abc");
		let result = identifier(input);
		assert!(result.is_err(), "Expected an error for identifier starting with a number");
	}

	#[test]
	fn test_identifier_mixed_case() {
		let input = Span::new("abcDEF123");
		let result = identifier(input);
		let (remaining, parsed) = result.expect("Failed to parse identifier");

		assert_eq!(parsed.fragment(), &"abcDEF123");
		assert_eq!(remaining.fragment(), &"");
	}

	#[test]
	fn test_identifier_at_end_of_input() {
		let input = Span::new("ABC_123\n");
		let result = identifier(input);
		let (remaining, parsed) = result.expect("Failed to parse identifier");

		assert_eq!(parsed.fragment(), &"ABC_123");
		assert_eq!(remaining.fragment(), &"\n");
	}

	#[test]
	fn test_identifier_longest_valid() {
		let input = Span::new("abc123_XYZ@invalid");
		let result = identifier(input);
		let (remaining, parsed) = result.expect("Failed to parse identifier");

		assert_eq!(parsed.fragment(), &"abc123_XYZ");
		assert_eq!(remaining.fragment(), &"@invalid");
	}

	#[test]
	fn test_value_separator_multiple_spaces_before() {
		let input = Span::new("   =   value");
		let result = value_separator(input);
		let (remaining, parsed) = result.expect("Failed to parse value separator");

		assert_eq!(parsed.fragment(), &"=");
		assert_eq!(remaining.fragment(), &"value");
	}

	#[test]
	fn test_value_separator_multiple_spaces_after() {
		let input = Span::new("= \t  ");
		let result = value_separator(input);
		let (remaining, parsed) = result.expect("Failed to parse value separator");

		assert_eq!(parsed.fragment(), &"=");
		assert_eq!(remaining.fragment(), &"");
	}

	#[test]
	fn test_value_separator_missing_equals() {
		let input = Span::new("KEY value");
		let result = value_separator(input);
		assert!(result.is_err(), "Expected an error for input without '=' separator");
	}

	#[test]
	fn test_value_separator_at_end_of_file() {
		let input = Span::new("=");
		let result = value_separator(input);
		let (remaining, parsed) = result.expect("Failed to parse value separator");

		assert_eq!(parsed.fragment(), &"=");
		assert_eq!(remaining.fragment(), &"");
	}

	#[test]
	fn test_values_comment() {
		let input = Span::new("# This is a comment");
		let result = values(input);
		let (remaining, parsed) = result.expect("Failed to parse values");

		assert_eq!(parsed.len(), 1);
		match &parsed[0] {
			ValuePart::Comment(comment) => assert_eq!(comment, " This is a comment"),
			_ => panic!("Expected a Comment ValuePart"),
		}
		assert_eq!(remaining.fragment(), &"");
	}

	#[test]
	fn test_values_variable_with_dollar_prefix() {
		let input = Span::new("$VAR_NAME");
		let result = values(input);
		let (remaining, parsed) = result.expect("Failed to parse values");

		assert_eq!(parsed.len(), 1);
		match &parsed[0] {
			ValuePart::Variable(var_name) => assert_eq!(var_name, "VAR_NAME"),
			_ => panic!("Expected a Variable ValuePart"),
		}
		assert_eq!(remaining.fragment(), &"");
	}

	#[test]
	fn test_values_variable_with_curly_braces() {
		let input = Span::new("${VAR_NAME}");
		let result = values(input);
		let (remaining, parsed) = result.expect("Failed to parse values");

		assert_eq!(parsed.len(), 1);
		match &parsed[0] {
			ValuePart::Variable(var_name) => assert_eq!(var_name, "VAR_NAME"),
			_ => panic!("Expected a Variable ValuePart"),
		}
		assert_eq!(remaining.fragment(), &"");
	}

	#[test]
	fn test_values_static_part() {
		let example_input = "THis is aosidnon409ansd0 9j0an oians**d6a7d98odin on";
		let input = Span::new(example_input);
		let result = values(input);
		let (remaining, parsed) = result.expect("Failed to parse values");

		assert_eq!(parsed.len(), 1);
		match &parsed[0] {
			ValuePart::Static(value) => assert_eq!(value, example_input),
			_ => panic!("Expected a Static ValuePart"),
		}
		assert_eq!(remaining.fragment(), &"");
	}

	#[test]
	fn test_values_multiple_parts() {
		let input = Span::new("static1 $VAR1 ${VAR2} static2 # Comment");
		let result = values(input);
		let (_, parsed) = result.expect("Failed to parse values");

		assert_eq!(
			&parsed,
			&[
				ValuePart::Static(String::from("static1 ")),
				ValuePart::Variable(String::from("VAR1")),
				ValuePart::Static(String::from(" ")),
				ValuePart::Variable(String::from("VAR2")),
				ValuePart::Static(String::from(" static2 ")),
				ValuePart::Comment(String::from(" Comment")),
			]
		);
	}

	#[test]
	fn test_values_only_whitespace() {
		let input = Span::new("   \t  ");
		let result = values(input);
		let (remaining, parsed) = result.expect("Failed to parse values");

		assert_eq!(parsed.len(), 1);
		match &parsed[0] {
			ValuePart::Static(value) => assert_eq!(value, "   \t  "),
			_ => panic!("Expected a Static ValuePart"),
		}
		assert_eq!(remaining.fragment(), &"");
	}

	#[test]
	fn test_values_consecutive_variables() {
		let input = Span::new("$VAR1${VAR2}$VAR3");
		let result = values(input);
		let (remaining, parsed) = result.expect("Failed to parse values");

		assert_eq!(
			parsed,
			vec![
				ValuePart::Variable("VAR1".to_string()),
				ValuePart::Variable("VAR2".to_string()),
				ValuePart::Variable("VAR3".to_string()),
			]
		);
		assert_eq!(remaining.fragment(), &"");
	}
	#[test]
	fn test_values_consecutive_variables_same_type() {
		let input = Span::new("$VAR1$VAR2$VAR3");
		let result = values(input);
		let (remaining, parsed) = result.expect("Failed to parse values");

		assert_eq!(
			parsed,
			vec![
				ValuePart::Variable("VAR1".to_string()),
				ValuePart::Variable("VAR2".to_string()),
				ValuePart::Variable("VAR3".to_string()),
			]
		);
		assert_eq!(remaining.fragment(), &"");
	}

	#[test]
	fn test_line_comment_with_whitespace() {
		let input = Span::new("  # This is a comment  \n");
		let result = line(input);
		let (remaining, parsed) = result.expect("Failed to parse line");

		match parsed {
			FileLine::Comment(comment) => assert_eq!(comment, " This is a comment  "),
			_ => panic!("Expected a Comment FileLine"),
		}
		assert_eq!(remaining.fragment(), &"\n");
	}

	#[test]
	fn test_line_key_value_pair_empty_value() {
		let input = Span::new("KEY=");
		let result = line(input);
		let (remaining, parsed) = result.expect("Failed to parse line");

		assert_eq!(remaining.fragment(), &"");
		match parsed {
			FileLine::KeyValue { key, value } => {
				assert_eq!(key, "KEY");
				assert!(value.is_empty());
			}
			_ => panic!("Expected KeyValue FileLine"),
		}
	}
	#[test]
	fn test_line_key_value_with_multiple_variables() {
		let input = Span::new("KEY = value_${VAR1}_${VAR2}_$VAR3");
		let result = line(input);
		let (remaining, parsed) = result.expect("Failed to parse line");

		assert_eq!(remaining.fragment(), &"");
		match parsed {
			FileLine::KeyValue { key, value } => {
				assert_eq!(key, "KEY");
				assert_eq!(value.len(), 6);
				assert_eq!(value[0], ValuePart::Static("value_".to_string()));
				assert_eq!(value[1], ValuePart::Variable("VAR1".to_string()));
				assert_eq!(value[2], ValuePart::Static("_".to_string()));
				assert_eq!(value[3], ValuePart::Variable("VAR2".to_string()));
				assert_eq!(value[4], ValuePart::Static("_".to_string()));
				assert_eq!(value[5], ValuePart::Variable("VAR3".to_string()));
			}
			other => panic!("Expected KeyValue FileLine, got {other:?}"),
		}
	}

	#[test]
	fn test_file_empty_input() {
		let input = Span::new("");
		let result = file(input);
		let (remaining, parsed) = result.expect("Failed to parse empty file");

		assert!(parsed.is_empty(), "Expected empty Vec for an empty input");
		assert_eq!(remaining.fragment(), &"", "Expected remaining input to be empty");
	}

	#[test]
	fn test_file_multiple_lines() {
		let input = Span::new("KEY1=value1\n# Comment\nKEY2=value2\n");
		let result = file(input);
		let (remaining, parsed) = result.expect("Failed to parse file");

		assert_eq!(remaining.fragment(), &"\n");
		assert_eq!(parsed.len(), 3);

		assert_eq!(
			&parsed[0],
			&FileLine::key_value("KEY1", vec![ValuePart::Static("value1".to_string())])
		);
		assert_eq!(&parsed[1], &FileLine::comment(" Comment"));
		assert_eq!(
			&parsed[2],
			&FileLine::key_value("KEY2", vec![ValuePart::Static("value2".to_string())])
		);
	}

	#[test]
	fn test_file_only_empty_lines() {
		let input = Span::new("\n\n\n");
		let result = file(input);
		let (remaining, parsed) = result.expect("Failed to parse file with only empty lines");

		assert_eq!(remaining.fragment(), &"\n");
		assert_eq!(parsed.len(), 3);

		assert_eq!(parsed, vec![FileLine::empty(), FileLine::empty(), FileLine::empty()]);
	}

	#[test]
	fn test_file_with_empty_line_in_middle() {
		let input = Span::new("KEY1=value1\n\nKEY2=value2");
		let result = file(input);
		let (remaining, parsed) = result.expect("Failed to parse file with empty line");

		assert_eq!(remaining.fragment(), &"");
		assert_eq!(parsed.len(), 3);

		assert_eq!(
			&parsed[0],
			&FileLine::key_value("KEY1", vec![ValuePart::Static("value1".to_string())])
		);
		assert_eq!(&parsed[1], &FileLine::empty());
		assert_eq!(
			&parsed[2],
			&FileLine::key_value("KEY2", vec![ValuePart::Static("value2".to_string())])
		);
	}

	#[test]
	fn test_file_only_comments() {
		let input = Span::new("# Comment 1\n# Comment 2\n# Comment 3");
		let result = file(input);
		let (remaining, parsed) = result.expect("Failed to parse file with only comments");

		assert_eq!(remaining.fragment(), &"");
		assert_eq!(parsed.len(), 3);

		assert_eq!(&parsed[0], &FileLine::comment(" Comment 1"));
		assert_eq!(&parsed[1], &FileLine::comment(" Comment 2"));
		assert_eq!(&parsed[2], &FileLine::comment(" Comment 3"));
	}

	#[test]
	fn test_file_with_duplicate_keys() {
		let input = Span::new("KEY1=value1\nKEY1=value2\nKEY2=value3\nKEY1=value4");
		let result = file(input);
		let (remaining, parsed) = result.expect("Failed to parse file with duplicate keys");

		assert_eq!(remaining.fragment(), &"");
		assert_eq!(parsed.len(), 4);

		assert_eq!(
			&parsed[0],
			&FileLine::key_value("KEY1", vec![ValuePart::Static("value1".to_string())])
		);
		assert_eq!(
			&parsed[1],
			&FileLine::key_value("KEY1", vec![ValuePart::Static("value2".to_string())])
		);
		assert_eq!(
			&parsed[2],
			&FileLine::key_value("KEY2", vec![ValuePart::Static("value3".to_string())])
		);
		assert_eq!(
			&parsed[3],
			&FileLine::key_value("KEY1", vec![ValuePart::Static("value4".to_string())])
		);
	}

	#[test]
	fn test_file_key_without_value() {
		let input = Span::new("KEY_ONLY=\n");
		let result = file(input);
		let (remaining, parsed) = result.expect("Failed to parse file");

		assert_eq!(remaining.fragment(), &"\n");
		assert_eq!(parsed.len(), 1);
		match &parsed[0] {
			FileLine::KeyValue { key, value } => {
				assert_eq!(key, "KEY_ONLY");
				assert!(value.is_empty(), "Expected empty value for key without equals sign");
			}
			_ => panic!("Expected KeyValue FileLine"),
		}
	}
	#[test]
	fn test_file_mixed_lines_with_empty_values() {
		let input = Span::new("KEY1=value1\nKEY2=\nKEY3=value3\nKEY4=");
		let result = file(input);
		let (remaining, parsed) = result.expect("Failed to parse file");

		assert_eq!(remaining.fragment(), &"");
		assert_eq!(parsed.len(), 4);

		assert_eq!(
			&parsed[0],
			&FileLine::key_value("KEY1", vec![ValuePart::Static("value1".to_string())])
		);
		assert_eq!(&parsed[1], &FileLine::key_value("KEY2", Vec::new()));
		assert_eq!(
			&parsed[2],
			&FileLine::key_value("KEY3", vec![ValuePart::Static("value3".to_string())])
		);
		assert_eq!(&parsed[3], &FileLine::key_value("KEY4", Vec::new()));
	}
}
